// =============== Klassen-Definition: MATRIX  ========================================
// MATRIX-Klasse fuer Matrizenoperationen vom Typ double
// Rechteckmatrix mat(i,j,k,l) wird spaltenweise gespeichert !!!
// Indizierung von i,j,k,l beginnt bei 0 !!!
// vgl. U. Breymann, C++, Carl Hanser Verlag 1993, S.186 ff
// ======================================================================================
// Copyright: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Erkrath/Darmstadt
// experimentelles Lehr- und Studienprogramm - kommerzielle Nutzung nicht gestattet
// Der Autor bernimmt keine Haftung fr Unvollstndigkeiten und Maengel der Algorithmen!
// Version:  5/1995
// Compiler: C++
// ================ Include-Files =======================================================
#include "pch.h"
#include "framework.h"
#include "MATRIXclass.h"
// ---------------------
// #include "Matrix.h"
// #include "Matrizen.h"
// ============================================================================
// public -> Oeffentliche Mitgliedsfunktionen
// ============================================================================
//
// Konstruktor -> MATRIX mat(ni,nj,nk,nl)
// ======================================
// ni, nj, nk, nl: Dimensionen der 4D-Matrix
// -----------------------------------------
MATRIX::MATRIX(int ni, int nj, int nk, int nl)
{
//*** std::cerr<< "MATRIX::Konstruktor: neue Matrix erzeugen" << '\n';
//
this->erzeugen(ni, nj, nk, nl);
};
//---------------------------------------------------------------------------
//
// Konstruktor -> MATRIX mat(C_array,ni,nj,nk,nl)
// ==============================================
// C_array vorinitialisieren mittels Initialisierungsliste C_feld [] = { ... }
// ni, nj, nk, nl: Dimensionen der 4D-Matrix
// --------------------------------------------------------------------------
MATRIX::MATRIX(double* C_array, int ni, int nj, int nk, int nl)
{
//*** std::cerr<< "MATRIX::Konstruktor: neue Matrix erzeugen" << '\n';
//
this->erzeugen(ni,nj,nk,nl,1,C_array);
};
//---------------------------------------------------------------------------
//
// Kopier-Konstruktor (Matrix vollstaendig kopieren) -> MATRIX mat1(&mat2)
// =======================================================================
MATRIX::MATRIX(MATRIX &mat2)
{
//*** std::cerr<<"MATRIX::Kopier-Konstruktor: Kopie von MATRIX erzeugen"<<'\n';
// 
// permanente Matrix
// =================
this->kopieren(mat2);
};//---------------------------------------------------------------------------
//
// Destruktor -> delete &mat
// =========================
MATRIX::~MATRIX(void)
{
//*** std::cerr<<"MATRIX::Destruktor: Matrix loeschen, typ= "<<typ<<'\n';
//
// Sperre gegen Loeschen beruecksichtigen
//
if(typ != 2)
{
if(typ != 0)
{
 if(matrix != NULL)
 { delete [] matrix;
   matrix=aktuelles_element=NULL;
 };
//*** std::cerr<<"MATRIX::Destruktor: Matrix-Inhalt geloescht"<<'\n';
};
//
typ=1;
dimi=dimj=dimk=diml=dimmax=0;
eintyp=0;
ia=ja=ka=la=0;
}
else
{
// Matrix darf nicht geloescht werden !!!
std::cout<<endl<<"MATRIX::Destruktor: *** Programmierfehler des Anwenders ***"<<endl;
std::cout<<endl<<"                    Matrix darf nicht geloescht werden !!! "<<endl;
std::cout<<endl<<"***********************************************************"<<endl;
};
//*** std::cerr<<"MATRIX::Destruktor: Matrix-Deskriptor geloescht"<<'\n';
};//---------------------------------------------------------------------------
//
// ============================================================================
// public -> Oeffentliche, virtuelle Mitgliedsfunktionen
// ============================================================================
//
// ============================================================================
// protected -> Geschuetzte Mitgliedsfunktionen
// ============================================================================
//
// Matrixelement -> mat.matrix_element(i,j,k,l) vergleiche mat(i,j,k,l)
// ====================================================================
double &MATRIX::matrix_element(int i,int j,int k,int l)
{
aktuelles_element=this->adresse(i,j,k,l);
return *aktuelles_element;
};//---------------------------------------------------------------------------
// ============================================================================
// protected -> Geschuetzte, virtuelle Mitgliedsfunktionen
// ============================================================================
//
// Adresse von Matrixelement -> mat.adresse(i,j,k,l)
// =================================================
double *MATRIX::adresse(int i,int j,int k,int l)
{
int ijkl;
char c;
// Vektor-Index ermitteln
ijkl=this->index(i,j,k,l); 
// Fehler abfangen
// ---------------
if(ijkl >= dimmax)
{
std::cout<<"*** Fehler in Funktion MATRIX::adresse -> Falscher Index:"<<'\n';
std::cout<<"dimmax="<<dimmax<<'\n';
std::cout<<"  ijkl="<<  ijkl<<'\n';
std::cin>>c;
return NULL;
}
//----------------
else
return this->matrix+ijkl;
};//---------------------------------------------------------------------------
//
// ============================================================================
// private -> Private, virtuelle Mitgliedsfunktionen
// ============================================================================
//
// Matrix erzeugen -> erzeugen(ni,nj,nk,nl,neutyp,C_array)
// =======================================================
// bei neutyp=-1 temporaere Matrix erzeugen
// bei neutyp= 1 permanente Matrix erzeugen
// C_array vorinitialisiertes C_Feld [] = { ... }
// ----------------------------------------
void MATRIX::erzeugen(int ni, int nj, int nk, int nl, int neutyp, double* C_array)
{
    // Matrix-Deskriptor belegen und Speicher reservieren
    // --------------------------------------------------
    //*** std::cerr<<"MATRIX::erzeugen: neutyp= "<<neutyp<<'\n';
    //
    // Typ belegen
    // -----------
    if (neutyp == 0) typ = -1;
    else            typ = neutyp;
    //
    // Vektor-Dimension belegen
    // ------------------------
    dimi = ni;
    dimj = nj;
    dimk = nk;
    diml = nl;
    dimmax = ni * nj * nk * nl;
    //
    // Speicher anfordern
    // ------------------
    matrix = new double[dimmax];
    //
    if (C_array == NULL)
    {
        // Matrix auf 0.0 setzen
        // ---------------------
        for (int i = 0; i < dimmax; i++) *(matrix + i) = 0.0;
    }
    else
    {
        // C_array einspeichern
        // --------------------
        for (int i = 0; i < dimmax; i++) *(matrix + i) = C_array[i];
    };
//
// aktuelles Matrix-Element
// ------------------------
aktuelles_element = matrix;
//
// Anfangsindizes belegen
// ----------------------
ia = ja = ka = la = 0;
//
// Typ des Einspeicherns belegen
// -----------------------------
eintyp = 1;
};//--------------------------------------------------------------------
//
// Matrix kopieren -> kopieren(&mat,neutyp)
// ========================================
// bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
// bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
// bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
// --------------------------------------------------------------------------
void MATRIX::kopieren(MATRIZEN &mat_,int neutyp)
{
// Referenttyp wandeln
MATRIX &mat=this->referenz(mat_);
// ------------------------------
//*** if(neutyp==-1) std::cerr<<"MATRIX::kopieren: nur MATRIX-Deskriptor kopieren, neuen Speicher anfordern"<<'\n';
//*** if(neutyp== 0) std::cerr<<"MATRIX::kopieren: nur MATRIX-Deskriptor kopieren, alten Speicher benutzen"<<'\n';
//*** if(neutyp>= 1) std::cerr<<"MATRIX::kopieren: MATRIX-Deskriptor kopieren, neuen Speicher anfordern, Inhalt kopieren"<<'\n';
//
// Matrixtyp
typ=neutyp;
// Matrixdimensionen
dimi=mat.dimi;
dimj=mat.dimj;
dimk=mat.dimk;
diml=mat.diml;
// Vektor-Dimension
dimmax=mat.dimmax;
// Typ des Einspeicherns
eintyp=mat.eintyp;
// Anfangsindizes
ia=mat.ia;
ja=mat.ja;
ka=mat.ka;
la=mat.la;
//*** std::cerr<<"MATRIX::kopieren: Matrix-Deskriptor kopiert"<<'\n';
//
if(neutyp == 0)
{
// nur Matrix-Adresse kopieren
matrix=mat.matrix;
aktuelles_element=matrix;
//*** std::cerr<<"MATRIX::kopieren: Matrix-Adresse kopiert"<<'\n';
}
else
{
// neuen Speicher anfordern
matrix=new double[dimmax];
aktuelles_element=matrix; 
//*** std::cerr<<"MATRIX::kopieren: neuer Speicher angefordert"<<'\n';
};
//
if(neutyp >= 1)
{
// Matrix-Inhalt kopieren
for(int i=0; i<dimmax; i++)
{
*(matrix+i)=*(mat.matrix+i);
};
//*** std::cerr<<"MATRIX::kopieren: Matrix-Inhalt kopiert"<<'\n';
}
else
if(neutyp == -1)
{
// Matrix-Inhalt auf null setzen
for(int i=0; i<dimmax; i++)
{
*(matrix+i)=0.0;
};
//*** std::cerr<<"MATRIX::kopieren: Matrix-Inhalt = 0.0"<<'\n';
};
};//---------------------------------------------------------------------------
//
// temporaere Matrix erzeugen -> temporaere_matrix_erzeugen(ni,nj,nk,nl,neutyp)
// ============================================================================
// bei neutyp=-1 temporaere Matrix erzeugen
// bei neutyp= 1 permanente Matrix erzeugen
// ni, nj, nk, nl: Dimensionen der 4D-Matrix
// -----------------------------------------
MATRIZEN *MATRIX::temporaere_matrix_erzeugen(int ni,int nj,int nk,int nl,int neutyp)
{
//*** std::cerr<<"MATRIX::temporaere_matrix_erzeugen: neue Matrix erzeugen"<<'\n'; 
//
MATRIZEN* temp = new MATRIX(ni, nj, nk, nl, neutyp);
//
return temp;
};//---------------------------------------------------------------------------
//
// temporaere Matrix als Kopie erzeugen -> temporaere_matrix_erzeugen(&mat,neutyp)
// ===============================================================================
// Kopie der Matrix mat:
// bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
// bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
// bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
// --------------------------------------------------------------------------
MATRIZEN *MATRIX::temporaere_matrix_erzeugen(MATRIZEN &mat_,int neutyp)
{
//*** std::cerr<<"MATRIX::temporaere_matrix_erzeugen: Kopie erzeugen: neutyp= "<<neutyp<<'\n';
// 
// Referenztyp wandeln
MATRIX &mat=this->referenz(mat_);
// ------------------------------
//
MATRIZEN *temp=new MATRIX(mat,neutyp);
//
return temp;
};//---------------------------------------------------------------------------
//
// Temporaere Matrix loeschen, -> mat.temporaere_matrix_loeschen()
// ===============================================================
void MATRIX::temporaere_matrix_loeschen()
{
// ***** Fehler behoben 10/2021: bei Multiplikation Dimensionen (nur Deskriptor) temporr ndern, aber Original anschlieend nichtt lschen   *****
//*** if(typ < 1)
// ************************************************************************************************************************************************
if (typ < 0)
{
// Temporaere Matrix loeschen
// --------------------------
//*** std::cerr<<"MATRIX::temporaere_matrix_loeschen, typ= "<<typ<<'\n';
delete this;
//*** std::cerr<<"MATRIX: temporaere Matrix geloescht"<<'\n';
};
};//---------------------------------------------------------------------------
//
// Matrix-Dimensionen temporaer aendern -> mat.dimension(i,j,k,l)
// ==============================================================
MATRIX &MATRIX::dimension(int ni,int nj,int nk,int nl)
{
// Zwischenspeicher temp
MATRIX *temp=new MATRIX(*this,0);
// ------------------------------
temp->dimension_temporaer(ni,nj,nk,nl);
// -----------------------------------------------
// ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
this->temporaere_matrix_loeschen();
//
return *temp;
};//--------------------------------------------------------------------
//
// Matrix-Dimensionen temporaer aendern -> mat.dimension_temporaer(i,j,k,l)
// ========================================================================
void MATRIX::dimension_temporaer(int ni,int nj,int nk,int nl)
{
char c;
int max,ijkl;
max=ni*nj*nk*nl;
if(max > dimmax)
{
std::cout<<"*** Fehler in Funktion MATRIX::dimension_temporaer -> Falsche Indizes:"<<'\n';
std::cout<<"dimmax="<<dimmax<<'\n';
std::cout<<"dimi="<<dimi<<" dimj="<<dimj<<" dimk="<<dimk<<" diml="<<diml<<'\n';
std::cout<<"  ni="<<  ni<<"   nj="<<  nj<<"   nk="<<  nk<<"   nl="<<  nl<<'\n';
std::cout<<"max   ="<<max<<'\n';
std::cin>>c;
};
// -----------------------------------
// Anfangsindex in der Originalmatrix
if(eintyp != 1) ijkl=index(ia,ja,ka,la);
// -------------------------------------
// Dimensionen aendern
dimi=ni;
dimj=nj;
dimk=nk;
diml=nl;
dimmax=max;
// --------------
// Anfangsindizes in temporaerer Matrix aendern
if(eintyp != 1) index_4d(ijkl,ia,ja,ka,la);
};//--------------------------------------------------------------------
//
// Teilmatrix heraus -> ...=mat1.teil_heraus(mat2, i,j,k,l)
// ========================================================
void MATRIX::teil_heraus(MATRIZEN &mat_, int i,int j,int k,int l)
{
// Referenttyp wandeln
MATRIX &mat=this->referenz(mat_);
// -----------------------------
// Anfang der Teilmatrix
int anfang=mat.index(i,j,k,l);
// -----------------------------
// Anzahl der Matrixelemente
int ende=anfang+(this->dimi)*(this->dimj)*(this->dimk)*(this->diml);
if(ende > mat.dimmax) ende=mat.dimmax;
// -----------------------------------------
// Matrixelemente kopieren
int q=0;
for(int p=anfang; p<ende; p++)
{
*(this->matrix+q)=*(mat.matrix+p);
q++;
};
};//---------------------------------------------------------------------------
//
// Teilmatrix hinein -> mat.teil_hinein(i,j,k,l)=...
// =================================================
void MATRIX::teil_hinein(int i,int j,int k,int l)
{
// Typ des Einspeicherns notieren
eintyp=0;
// Anfangsindizes notieren
ia=i;
ja=j;
ka=k;
la=l;
};//--------------------------------------------------------------------
//
// Untermatrix heraus -> ...=mat1.unter_heraus(mat2, i,j,k,l)
// ==========================================================
void MATRIX::unter_heraus(MATRIZEN &mat_, int i,int j,int k,int l)
{
// Referenttyp wandeln
MATRIX &mat=this->referenz(mat_);
// ------------------------------
int ii,jj,kk,ll;
//
for (ll=0; ll<diml; ll++)
{                                      // Anfang l
  for (kk=0; kk<dimk; kk++)
  {                                    // Anfang k
    for (jj=0; jj<dimj; jj++)
    {                                  // Anfang j
      for (ii=0; ii<dimi; ii++)
      {                                // Anfang i
         this->matrix_element(ii,jj,kk,ll)=mat.matrix_element(i+ii,j+jj,k+kk,l+ll);
      };                               // Ende i
    };                                 // Ende j
  };                                   // Ende k
};                                     // Ende l
};//--------------------------------------------------------------------
//
// Untermatrix hinein -> mat.unter_hinein(i,j,k,l)=...
// ===================================================
void MATRIX::unter_hinein(int i,int j,int k,int l)
{
// Typ des Einspeicherns notieren
eintyp=-1;
// Anfangsindizes notieren
ia=i;
ja=j;
ka=k;
la=l;
};//--------------------------------------------------------------------
//
// Matrix-Ausgabe -> mat.ausgeben(s)
// =================================
void MATRIX::ausgeben(std::ostream& s)
{
int const spalten=7;
int k,ek,ik,ei,i,dk,eik,ii,jj,kk,ll,dij,dkl;
// -----------------------------------------------
// Ueberschrift ausgeben
if(&s == &std::cout)
{
std::cout<<endl;
std::cout<<"[ ("<<dimi<<"*"<<dimj<<") * ("<<dimk<<"*"<<diml;
std::cout<<") ]-MATRIX:";
std::cout<<endl;
};
// =====================================================================
// Matrix-Deskriptor fuer Testzwecke drucken
// -----------------------------------------
//*** std::cerr<<"MATRIX::ausgeben - ";
//*** std::cerr<<"Adresse= "<<this<<'\n';
//*** std::cerr<<"typ, dimi, dimj, dimk, diml, dimmax= "<<typ<<' '<<dimi<<' '<< dimj<<' '<<dimk<<' '<<diml<<' '<<dimmax<<'\n';
//*** std::cerr<<"eintyp, ia, ja, ka, la, matrix, aktuelles_element= "<<eintyp<<' '<<ia<<' '<< ja<<' '<<ka<<' '<<la<<' '<<matrix<<' '<<aktuelles_element;
//*** std::cerr<<'\n';
// =====================================================================
dkl=dimk*diml;
dij=dimi*dimj;
//
if(dkl == 1)
{
ek=dimj;
ei=dimi;
}
else
if(dij == 1)
{
ek=diml;
ei=dimk;
}
else
{
ek=dimk*diml;
ei=dimi*dimj;
};
//
for (k=0; k < ek; k+=spalten)
{
dk=k+spalten;
if (dk > ek) dk=ek;
if(&s == &std::cout) std::cout<<"Spalten "<<k<<" bis "<<dk-1<<'\n';
//*************************************************************
// 2D-Matrix mat(i, ik) ausgeben, Dimensionen mat(ni*nj, nk*nl)
//*************************************************************
  for (i=0; i < ei; i++)
  {
    eik=k+spalten;
    if(eik > ek) eik=ek;
// Indizes ii,jj aus i
    if(dkl == 1) goto Marke;
    if(dij == 1) goto Marke;
    index_2d(i, dimi,dimj, ii,jj);
//
Marke:
     for (ik=k; ik < eik; ik++)
     {
     // Feldbreite
     // ----------
     if(&s == &std::cout) std::cout.width(10);
     //-----------
if(dkl == 1)
{
    s<<matrix_element(i,ik,0,0)<<" ";
}
else
if(dij == 1)
{
    s<<matrix_element(0,0,i,ik);
}
else
{
// Indizes kk,ll aus ik berechnen
    index_2d(ik, dimk,diml, kk,ll);
//
    s<<matrix_element(ii,jj,kk,ll)<<" ";
};
    };
  s<<endl;
  };
if(dk != ek) s<<endl;
};
std::cout << endl;
};//----------------------------------------------------------------------------
//
// Matrix-Eingabe -> mat.einlesen(s)
// =================================
void MATRIX::einlesen(std::istream &s)
{
int i,j,k,l;
// ---------
if(&s == &std::cin)
{
  std::cout<<" [ ("<<dimi<<"*"<<dimj<<") * ("<<dimk<<"*"<<diml;
  std::cout<<") ]-MATRIX (zeilenweise !!!) eingeben:";
  // bel, bel
  std::cout<<'\07'<<'\n';
};
// -------------------
for (i=0; i<dimi; i++)
for (j=0; j<dimj; j++)
for (k=0; k<dimk; k++)
for (l=0; l<diml; l++)
s>>matrix_element(i,j,k,l);
};//--------------------------------------------------------------------
//
// Alle Matrix-Elemente belegen -> mat =zahl
// =========================================
// Hier nochmal, weil der Zuweisungsoperator  = nicht vererbt werden kann !!!
// --------------------------------------------------------------------------
MATRIX &MATRIX::operator = (double zahl)
{
this->zuweisen(zahl,0);
//
// ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
this->temporaere_matrix_loeschen();
//
return *this;
};//--------------------------------------------------------------------
//
// Matrix zuweisen -> mat1 = mat2 (nur Matrizeninhalt kopieren)
// ==============================
// Hier nochmal, weil der Zuweisungsoperator  = nicht vererbt werden kann !!!
// --------------------------------------------------------------------------
MATRIX &MATRIX::operator  = (MATRIZEN &mat2)
{
this->zuweisen(mat2,0);
//
// ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
this->temporaere_matrix_loeschen();
mat2.temporaere_matrix_loeschen();
//
return *this;
};//---------------------------------------------------------------------------
//
// Alle Matrix-Elemente belegen -> mat.zuweisen(zahl,zweig)
//                                 (nur Matrizeninhalt kopieren)
// =============================================================
// Zweig= 0 -> mat=    zahl
//      = 2 -> mat=mat+zahl
//      = 4 -> mat=mat-zahl
// ------------------------
void MATRIX::zuweisen(double zahl,int zweig)
{
int i,k;
double *matrix1;
//
if(eintyp == -1)
  // Untermatrix
{
  if(zweig == 0) matrix_element(ia,ja,ka,la) =zahl;
    else
  if(zweig == 2) matrix_element(ia,ja,ka,la)+=zahl;
    else
  if(zweig == 4) matrix_element(ia,ja,ka,la)-=zahl;
  }
else
{
  // Matrix und Teilmatrix
  k=0;
  if(eintyp == 0) k=index(ia,ja,ka,la);
  matrix1=matrix+k;
  //
  if(zweig == 0) for (i=k; i<dimmax ;i++) *(matrix1++) =zahl;
  else
  if(zweig == 2) for (i=k; i<dimmax ;i++) *(matrix1++)+=zahl;
  else
  if(zweig == 4) for (i=k; i<dimmax ;i++) *(matrix1++)-=zahl;
};
};//------------------------------------------------------------------
//
// Matrix zuweisen -> mat1.zuweisen(mat2,zweig)
//                    (nur Matrizeninhalt kopieren)
// ================================================
// Zweig= 0 -> mat1=    mat2
//      = 2 -> mat1=mat+mat2
//      = 4 -> mat1=mat-mat2
void MATRIX::zuweisen(MATRIZEN &mat2_,int zweig)
{
// Referenztyp wandeln
MATRIX &mat2=this->referenz(mat2_);
// --------------------------------
int i,j,k,l,dim;
double *matrix1,*matrix2;
//
if(eintyp == -1)
  // Untermatrix
{
  if(zweig == 0)
    for (l=0; l<mat2.diml; l++)
      for (k=0; k<mat2.dimk; k++)
        for (j=0; j<mat2.dimj; j++)
          for (i=0; i<mat2.dimi; i++)
              this->matrix_element(ia+i,ja+j,ka+k,la+l) = mat2.matrix_element(i,j,k,l);
  else
  if(zweig == 2)
    for (l=0; l<mat2.diml; l++)
      for (k=0; k<mat2.dimk; k++)
        for (j=0; j<mat2.dimj; j++)
          for (i=0; i<mat2.dimi; i++)
              this->matrix_element(ia+i,ja+j,ka+k,la+l)+=mat2.matrix_element(i,j,k,l);
  else
  if(zweig == 4)
    for (l=0; l<mat2.diml; l++)
      for (k=0; k<mat2.dimk; k++)
        for (j=0; j<mat2.dimj; j++)
          for (i=0; i<mat2.dimi; i++)
              this->matrix_element(ia+i,ja+j,ka+k,la+l)-=mat2.matrix_element(i,j,k,l);
  }
else
  {
    // Teilmatrix
    k=0;
    if(eintyp == 0) k=index(ia,ja,ka,la);
    // Einspeichern von Matrix und Teil-Matrix
  // Indexgrenzen
    dim=dimmax-k;
    if(mat2.dimmax < dim) dim=mat2.dimmax;
    matrix1=matrix+k;
    matrix2=mat2.matrix;
    if(zweig == 0)
      for (matrix2=mat2.matrix; matrix2<mat2.matrix+dim ;matrix2++)
         *(matrix1++) =*matrix2;
    else
    if(zweig == 2)
      for (matrix2=mat2.matrix; matrix2<mat2.matrix+dim ;matrix2++)
         *(matrix1++)+=*matrix2;
    else
    if(zweig == 4)
      for (matrix2=mat2.matrix; matrix2<mat2.matrix+dim ;matrix2++)
         *(matrix1++)-=*matrix2;
  };
};//------------------------------------------------------------------
//
// Addition -> mat1+mat2 bzw. mat1+vek2 bzw. vek1+mat2
// ===================================================
void MATRIX::addieren(MATRIZEN &mat1_,MATRIZEN &mat2_)
{
// Referenztyp wandeln
MATRIX &mat1=this->referenz(mat1_);
MATRIX &mat2=this->referenz(mat2_);
// --------------------------------
int i,k,dim;
// ---------
if (mat2.dimmax > mat1.dimmax)
{
dim=mat2.dimmax;
// -------------
k=0;
for (i=0; i<dim; i++)
{
*(this->matrix+i)=*(mat1.matrix+k)+*(mat2.matrix+i);
k++;
if(k==mat1.dimmax) k=0;
};
}
else
{
dim=mat1.dimmax;
// -------------
k=0;
for (i=0; i<dim; i++)
{
*(this->matrix+i)=*(mat1.matrix+i)+*(mat2.matrix+k);
k++;
if(k==mat2.dimmax) k=0;
};
};
};//--------------------------------------------------------------------
//
// Subtraktion -> mat1-mat2 bzw. mat1-vek2 bzw. vek1-mat2
// ======================================================
void MATRIX::subtrahieren(MATRIZEN &mat1_,MATRIZEN &mat2_)
{
// Referenztyp wandeln
MATRIX &mat1=this->referenz(mat1_);
MATRIX &mat2=this->referenz(mat2_);
// --------------------------------
int i,k,dim;
// ---------
if (mat2.dimmax > mat1.dimmax)
{
dim=mat2.dimmax;
// -------------
k=0;
for (i=0; i<dim; i++)
{
*(this->matrix+i)=*(mat1.matrix+k)-*(mat2.matrix+i);
k++;
if(k==mat1.dimmax) k=0;
};
}
else
{
dim=mat1.dimmax;
// -------------
k=0;
for (i=0; i<dim; i++)
{
*(this->matrix+i)=*(mat1.matrix+i)-*(mat2.matrix+k);
k++;
if(k==mat2.dimmax) k=0;
};
};
};//--------------------------------------------------------------------
//
// Matrizenmultiplikation -> matc(i,j,k,l) = mat1(i,s,j) * mat2(k,s,l)
// zweig = 0 -> matc = mat1 * mat2 / = 1 -> matc = matc + mat1 * mat2 / = -1 -> matc = matc - mat1 * mat2 
// ======================================================================================================
void MATRIX::multiplizieren(MATRIZEN& mat1_, MATRIZEN& mat2_, int zweig)
{
    bool FEHLER;
    FEHLER = false;
    //*** FEHLER = true;
    // Referenztyp wandeln
    MATRIX& mat1 = this->referenz(mat1_);
    //*** std::cout << "mat1 = ";
    //*** mat1.ausgeben(std::cout);
    //
    MATRIX& mat2 = this->referenz(mat2_);
    //*** std::cout << "mat2 = ";
    //*** mat2.ausgeben(std::cout);
    //
    //*** std::cerr << "matc: dimi = " << dimi << " dimj = " << dimj << " dimk = " << dimk << " diml =" << diml<< '\n';
    // --------------------------------
    // Indizes: Spalten-Anfang, -Abstand, -Ende der Matrizen
    double sum;
    char stop = 0;
    // Matrix C
    int a_ic = 0;
    // Matrix 1
    int ai, ei;
    int is, as, es, ds, dims;
    int ej, dj;
    // Matrix 2
    int ak, ek;
    int is_2, as_2, ds_2;
    int el, dl;
    //
        // falls Summationsindex von mat1 grer als von mat2, dann Summation krzen
        // ==========================================================================
    dims = mat1.dimj;
    if (dims > mat2.dimj) dims = mat2.dimj;
    // Warnmeldung
    if (mat1.dimj != mat2.dimj)
    {
        std::cerr << " *** Warnung fuer Multiplikation matc(i,j,k,l) = mat1(i,s,j) * mat2(k,s,l): " << endl;
        std::cerr << "     Summationsindex mat1.dims = " << mat1.dimj << " ungleich zu mat2.dims = " << mat2.dimj << " !!! " << endl;
        std::cerr << "     Summationsindex entsprechend gekuerzt auf dims = " << dims << " !!! " << endl;
    };
  //
  // sichere Test-Version *******************************************************************************
  FEHLER = false;
  //*** FEHLER = true;
  if (FEHLER == true)
  {
      std::cerr << "*** Multiplikation mit sicherer Testversion *** " << endl;
        int i, j, k, l, s;
        // Schleifen ber Ergebnismatrix
        for (l = 0; l < mat2.dimk; l++)
        {                                      // Anfang l
            for (k = 0; k < mat2.dimi; k++)
            {                                    // Anfang k
                for (j = 0; j < mat1.dimk; j++)
                {                                  // Anfang j
                    for (i = 0; i < mat1.dimi; i++)
                    {                                // Anfang i
                        sum = 0.0;
                        if (zweig != 0)
                        {
                            sum = this->matrix_element(i, j, k, l);
                            if (zweig == -1) sum = -sum;
                      };
                        // falls Summationsindex von mat1 kleiner als von mat2, dann mit dims < mat1.dimj Summation krzen
                        // ===============================================================================================
                        // innere Summationsschleife
                        for (s = 0; s < dims; s++)
                        {                              // Anfang s
                            sum = sum + mat1.matrix_element(i, s, j) * mat2.matrix_element(k, s, l);
                        };                             // Ende s
                        if (zweig == -1) sum = -sum;
                        this->matrix_element(i, j, k, l) = sum;
                    };                                // Ende i
                };                                 // Ende j
            };                                   // Ende k
        };                                     // End l
        //*** std::cerr << "Ergebnis der Multiplikation ";
        //*** std::cerr << "matc = ";
        //*** this->ausgeben(std::cout);
        //*** std::cerr << "Test-PAUSE: " << endl;
        //*** std::cin >> stop;
  }
  else
  {  
  // Optimierung der Testversion ********************************************************************************    
    //
    //Schleifen ber Ergebnismatrix -------------------------------------------------------
    ds_2 = mat2.dimi;                               // entspricht ds von mat1
    dl = mat2.dimi * mat2.dimj;
    el = mat2.dimi * mat2.dimj * mat2.dimk;
    //
    ds = mat1.dimi;
    dj = mat1.dimi * mat1.dimj;
    ej = mat1.dimi * mat1.dimj * mat1.dimk;
    //
    for (ak = 0; ak < el; ak = ak + dl)
        //*** for (l = 0; l < mat2.dimk; l++)
    {                                      // Anfang l
        //*** ak = mat2.dimi * mat2.dimj * l;
        ek = ak + ds_2;
        //------------------------------------------------------------------------------------
        for (as_2 = ak; as_2 < ek; as_2 = as_2 + 1)
            //*** for (k = 0; k < mat2.dimi; k++)
        {                                    // Anfang k
        //*** as_2 = k + mat2.dimi * mat2.dimj * l;
        //
            //*** ej = mat1.dimi * mat1.dimj * mat1.dimk;
            //*** dj = mat1.dimi * mat1.dimj;
       //------------------------------------------------------------------------------
            for (ai = 0; ai < ej; ai = ai + dj)
                //*** for (j = 0; j < mat1.dimk; j++)    
            {                                  // Anfang j
            //*** ai = mat1.dimi * mat1.dimj * j;
                ei = ai + ds;
                //*** ei = ai + mat1.dimi;
                //*** ds = mat1.dimi;
                // -------------------------------------------------------------------------
                for (as = ai; as < ei; as++)
                    //***     for (i = 0; i < mat1.dimi; i++)
                {                                // Anfang i
                //*** as = i  + mat1.dimi * mat1.dimj * j;
                //*** es = as + mat1.dimi * mat1.dimj;
                    es = as + dj;
                    //
                    // falls Summationsindex von mat1 grer als von mat2, dann Summation krzen
                    // =========================================================================
                    if (mat1.dimj > mat2.dimj) es = as + mat1.dimi * dims;
                    //
                    is_2 = as_2;
                    sum = 0.0;
                    if (zweig != 0)
                    {
                        sum = *(this->matrix + a_ic);
                        if (zweig == -1) sum = -sum;
                    };
                    // innere Summationsschleife -------------------------------------------------------------------------------------------------------------
                    for (is = as; is < es; is = is + ds)                              //  Angangsindizes von mat1 und mat2
                    {                               // Anfang s                       //  index_1 = i + mat1.dimi * s + mat1.dimi * mat1.dimj * j;
                                                                                      //  index_2 = k + mat2.dimi * s + mat2.dimi * mat2.dimj * l;
                        sum = sum + *(mat1.matrix + is) * *(mat2.matrix + is_2);
                        is_2 = is_2 + ds_2;
                    };                            // Ende s
                    if (zweig == -1) sum = -sum;
                    *(this->matrix + a_ic) = sum;
                    a_ic++;
                };                              // Ende i
            };                                 // Ende j
        };                                   // Ende k
    };                                    // Ende l
  };
}; //--------------------------------------------------------------------
//
// Matrizendivision -> matx(s,n) = matr(i,n) / MATk(i,s)
//         MATk(i,s) * matx(s,n) = matr(i,n)
// =====================================================
MATRIZEN& MATRIX::dividieren(MATRIZEN& matr_)
{
    bool FEHLER;
    FEHLER = false;
    //*** FEHLER = true;
    // ----------------------------------
    // Referenztyp wandeln
    MATRIX& matr = this->referenz(matr_);
    // ----------------------------------
    std::cout << "MATRIX::dividieren: => Fehler: die Koeffizientenmatrix muss vom TYP qMATRIX sein !!! " << endl;
    std::cout << "Daher ergibt sich hier als Ergebnis: Loesungen := Rechte Seiten !!! " << endl;
    //
    // Ergebnisspeicher matx
    MATRIZEN* temp = temporaere_matrix_erzeugen(matr, 1);
    // ------------------------------------------------
    //
    // ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
    matr_.temporaere_matrix_loeschen();
    this->temporaere_matrix_loeschen();
    //
    return *temp;
}//---------------------------------------------------------------------------
//
// Matrix transponieren -> mat.transponieren(index_0,index_1,index_2,index_3)
// ==========================================================================
MATRIZEN &MATRIX::transponieren(int index_0,int index_1,int index_2,int index_3)
{
// Indexfeld mit den zu vertauschenden Indizes
int findex[4] = {0,0,0,0};
// ----------
// Zwischenspeicher temp
MATRIX *temp=new MATRIX(*this,-1);
// -------------------------------
// Standardfall abfragen (vorbelegte Parameter -> Aufruf ohne Parameter)
if(index_0 == 2 && index_1 == 3 && index_2 == 0 && index_3 ==1)
{
// 2D-Matrizen
// ===========
//
if((dimi*dimj == 1) || (dimk*diml == 1))
  {index_0=1; index_1=0; index_2=3; index_3=2;}
else
if((dimi*dimk == 1) || (dimj*diml == 1))
  {index_0=2; index_1=3; index_2=0; index_3=1;} 
else
if((dimi*diml == 1) || (dimj*dimk == 1))
  {index_0=3; index_1=2; index_2=1; index_3=0;};
};
// ---------------------------------------------
//Dimensionen tauschen
findex[0]=dimi;              findex[1]=dimj;
findex[2]=dimk;              findex[3]=diml;
temp->dimi=findex[index_0];  temp->dimj=findex[index_1];
temp->dimk=findex[index_2];  temp->diml=findex[index_3];
// -----------------------------------------------
int i,j,k,l;
//
for (l=0; l<diml; l++)
{ findex[3]=l;
  for (k=0; k<dimk; k++)
  { findex[2]=k;
    for (j=0; j<dimj; j++)
    { findex[1]=j;
      for (i=0; i<dimi; i++)
      { findex[0]=i;
     temp->matrix_element(findex[index_0],findex[index_1],findex[index_2],findex[index_3])=
           this->matrix_element(i,j,k,l);
      };
    };
  };
};
//
// ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
this->temporaere_matrix_loeschen();
//
return *temp;
};//--------------------------------------------------------------------
//
// Verjuengung einer Matrix -> mat.verjuengen(index_0,index_1,index_2,index_3)
// ===========================================================================
MATRIZEN &MATRIX::verjuengen(int index_0,int index_1,int index_2,int index_3)
{
int i0,i1,i2,i3,dim0,dim1,dim2,dim3;
// Indexfeld fuer zu bearbeitende Dimensionen 
int findex[4] = { 0,0,0,0 };
// Zwischenspeicher
MATRIX *temp=NULL;
// ----------
// Standardfall abfragen (vorbelegte Parameter -> Aufruf ohne Parameter)
if(index_0 == 0 && index_1 == 1 && index_2 == 2 && index_3 ==1)
{
// 2D-Matrizen
// ===========
if((dimi*dimj == 1) || (dimk*diml == 1))
  {index_0=0; index_1=0; index_2=2; index_3=2;}
else
if((dimi*dimk == 1) || (dimj*diml == 1))
  {index_0=0; index_1=1; index_2=0; index_3=1;} 
else
if((dimi*diml == 1) || (dimj*dimk == 1))
  {index_0=0; index_1=1; index_2=1; index_3=0;};
};
// ---------------------------------------------
findex[0]=dimi;
findex[1]=dimj;
findex[2]=dimk;
findex[3]=diml;
//
dim0=findex[index_0];
dim1=findex[index_1];
dim2=findex[index_2];
dim3=findex[index_3];
// -----------------
// Abpruefen, ob Spur gebildet wird
if(((index_0 == index_1) && (index_2 == index_3)) ||
   ((index_0 == index_2) && (index_1 == index_3)) ||
   ((index_0 == index_3) && (index_1 == index_2)))
{ // Spur bilden
  // ===========
  // Zwischenspeicher temp
  temp=new MATRIX(1,1,1,1,-1);
  // ---------------------------------
  temp->matrix_element(0)=this->spur(index_0,index_1,index_2,index_3);
  goto Ende;
}
else
{ // verjuengte Matrix bilden
  // ========================
  if(index_0 == index_1)
    {
      // mat(i,i,k,l)
      // ------------
      // Zwischenspeicher
      temp=new MATRIX(dim2,dim3,1,1,-1);
      // -------------------------------
      for(i3=0; i3<dim3; i3++)
      for(i2=0; i2<dim2; i2++)
      if(index_2 == 2)
      for(i0=0; i0<dim0; i0++) temp->matrix_element(i2,i3)+=this->matrix_element(i0,i0,i2,i3);
      else
      for(i0=0; i0<dim0; i0++) temp->matrix_element(i2,i3)+=this->matrix_element(i0,i0,i3,i2);
    }
  else
  if(index_0 == index_2)
    {
      // mat(i,j,i,l)
      // ------------
      // Zwischenspeicher
      temp=new MATRIX(dim1,dim3,1,1,-1);
      // -------------------------------
      for(i3=0; i3<dim3; i3++)
      for(i1=0; i1<dim1; i1++)
      if(index_1 == 1)
      for(i0=0; i0<dim0; i0++) temp->matrix_element(i1,i3)+=this->matrix_element(i0,i1,i0,i3);
      else
      for(i0=0; i0<dim0; i0++) temp->matrix_element(i1,i3)+=this->matrix_element(i0,i3,i0,i1);
    }
  else
  if(index_0 == index_3)
    {
      // mat(i,j,k,i)
      // ------------
      // Zwischenspeicher
      temp=new MATRIX(dim1,dim2,1,1,-1);
      // -------------------------------
      for(i2=0; i2<dim2; i2++)
      for(i1=0; i1<dim1; i1++)
      if(index_1 == 1)
      for(i0=0; i0<dim0; i0++) temp->matrix_element(i1,i2)+=this->matrix_element(i0,i1,i2,i0);
      else
      for(i0=0; i0<dim0; i0++) temp->matrix_element(i1,i2)+=this->matrix_element(i0,i2,i1,i0);
    }
  else
  if(index_1 == index_2)
    {
      // mat(i,j,j,l)
      // ------------
      // Zwischenspeicher
      temp=new MATRIX(dim0,dim3,1,1,-1);
      // -------------------------------
      for(i3=0; i3<dim3; i3++)
      for(i0=0; i0<dim0; i0++)
      if(index_0 == 0)
      for(i1=0; i1<dim1; i1++) temp->matrix_element(i0,i3)+=this->matrix_element(i0,i1,i1,i3);
      else
      for(i1=0; i1<dim1; i1++) temp->matrix_element(i0,i3)+=this->matrix_element(i3,i1,i1,i0);
    }
  else
  if(index_1 == index_3)
    {
      // Standardfall
      // mat(i,j,k,j)
      // ------------
      // Zwischenspeicher
      temp=new MATRIX(dim0,dim2,1,1,-1);
      // -------------------------------
      for(i2=0; i2<dim2; i2++)
      for(i0=0; i0<dim0; i0++)
      if(index_0 == 0)
      for(i1=0; i1<dim1; i1++) temp->matrix_element(i0,i2)+=this->matrix_element(i0,i1,i2,i1);
      else
      for(i1=0; i1<dim1; i1++) temp->matrix_element(i0,i2)+=this->matrix_element(i2,i1,i0,i1);
    }
  else
  if(index_2 == index_3)
    {
      // mat(i,j,k,k)
      // ------------
      // Zwischenspeicher
      temp=new MATRIX(dim0,dim1,1,1,-1);
      // -------------------------------
      for(i1=0; i1<dim1; i1++)
      for(i0=0; i0<dim0; i0++)
      if(index_0 == 0)
      for(i2=0; i2<dim2; i2++) temp->matrix_element(i0,i1)+=this->matrix_element(i0,i1,i2,i2);
      else
      for(i2=0; i2<dim2; i2++) temp->matrix_element(i0,i1)+=this->matrix_element(i1,i0,i2,i2);
    };
};
//
Ende:
// ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
this->temporaere_matrix_loeschen();
//
return *temp;
};//--------------------------------------------------------------------
//
// Spur einer Matrix -> mat.spur(index_0,index_1,index_2,index_3)
// ==============================================================
double MATRIX::spur(int index_0,int index_1,int index_2,int index_3)
{
int i0,i1,dim0,dim1;
double spur=0.0;
// Indexfeld fuer zu bearbeitende Dimensionen 
int findex[4] = { 0,0,0,0 };
// -------------
// Standardfall abfragen (vorbelegte Parameter -> Aufruf ohne Parameter)
if(index_0 == 0 && index_1 == 1 && index_2 == 0 && index_3 ==1)
{
// 2D-Matrizen
// ===========
if((dimi*dimj == 1) || (dimk*diml == 1))
  {index_0=0; index_1=0; index_2=2; index_3=2;}
else
if((dimi*dimk == 1) || (dimj*diml == 1))
  {index_0=0; index_1=1; index_2=0; index_3=1;} 
else
if((dimi*diml == 1) || (dimj*dimk == 1))
  {index_0=0; index_1=1; index_2=1; index_3=0;};
};
// ---------------------------------------------
findex[0]=dimi;
findex[1]=dimj;
findex[2]=dimk;
findex[3]=diml;
// ------------
if((index_0 == index_2) && (index_1 == index_3))
{
// Standardfall
// mat(i,j,i,j)
// ------------
   dim0=findex[index_0];
   dim1=findex[index_1];
   if((dim0 == findex[index_2]) && (dim1 == findex[index_3]))
   {
     for(i0=0; i0<dim0; i0++)
     for(i1=0; i1<dim1; i1++)
     spur=spur+(*this).matrix_element(i0,i1,i0,i1);
   };
}
else
if((index_0 == index_1) && (index_2 == index_3))
{
// mat(i,i,j,j)
// ------------
   dim0=findex[index_0];
   dim1=findex[index_2];
   if((dim0 == findex[index_1]) && (dim1 == findex[index_3]))
   {
     for(i0=0; i0<dim0; i0++)
     for(i1=0; i1<dim1; i1++)
     spur=spur+(*this).matrix_element(i0,i0,i1,i1);
   };
}
else
if((index_0 == index_3) && (index_1 == index_2))
{
// mat(i,j,j,i)
// ------------
   dim0=findex[index_0];
   dim1=findex[index_1];
   if((dim0 == findex[index_3]) && (dim1 == findex[index_2]))
   {
     for(i0=0; i0<dim0; i0++)
     for(i1=0; i1<dim1; i1++)
     spur=spur+(*this).matrix_element(i0,i1,i1,i0);
   };
};
// ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
this->temporaere_matrix_loeschen();
//
return spur;
};//--------------------------------------------------------------------
//
// Determinante von (3 x 3)- / (2 x 2)- / (1 x 1)-Matrizen bestimmen
// -> Amat.determinante(n)
// =================================================================
// n = 3 / 2 / 1 Dimension der Matrix  
// Ergebnis: Determinante der Matrix Amat
// -----------------------------------------------------------------
// Urheberrecht: Prof. Dr.-Ing. Udo F. Meissner, Erkrath
// Version:  11.11.2022 
// Compiler: C++ 
// -----------------------------------------------------------------
double MATRIX::determinante(int n)
{
    // Determinante g der Metrik Amat emitteln
    // =======================================
double g = 0.0;
if (n > 3)
{
    cout << " Determinante: Dimension zu gross = " << n << endl;
    return g;
}
// -----------------------------------------------------------------
MATRIX& mat = *this;
//
if (n == 1)
    {
        // 1D: det = A(I,0) * A(I,0)
        // =========================
        g = mat(0, 0);
    }
    else
    {
        if (n == 2)
        { // 2D: det = 1/2 * e(I,K)*A(I,M) * A(KN)*e(M,N)
          // ============================================
            MATRIX& e2 = *new e_MATRIX(2, 2);
            MATRIX& H1 = *new MATRIX(2, 2);
            //
            H1 = e2.dimension(1, 2, 2) * mat.dimension(1, 2, 2);
            MATRIX& H2 = *new MATRIX(2, 2);
            H2 = mat * e2;
            //
            g = 0.5 * H1.dimension(1, 2 * 2) * H2.dimension(1, 2 * 2);
            //
            delete& e2;
            delete& H1; delete& H2;
        }
        else
        {
            // 3D: det = 1/6 * e(I,K,L)*A(I,M) * A(L,P)*e(M,N,P) * A(K,N)
            // ========================================================== 
            MATRIX& e3 = *new e_MATRIX(3, 3, 3);
            MATRIX& H1 = *new MATRIX(3, 3, 3);
            //
            H1 = e3.dimension(1, 3, 3 * 3) * mat.dimension(1, 3, 3);
            MATRIX& H2 = *new MATRIX(3, 3, 3);
            //
            H2 = mat * e3.dimension(3 * 3, 3, 1);
            // 
            H1 = H1.dimension(3, 3 * 3, 1) * H2.dimension(1, 3 * 3, 3);
            //
            g = 1.0 / 6.0 * H1.dimension(1, 3 * 3, 1) * mat.dimension(1, 3 * 3, 1);
            delete& e3;
            delete& H1; delete& H2;
        }
     }
    //
// ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
this->temporaere_matrix_loeschen();
//
return g;
};// -------------------------------------------------------------------
//
// ============================================================================
// private -> Private Mitgliedsfunktionen
// ============================================================================
//
// Konstruktor -> MATRIX mat(ni,nj,nk,nl,neutyp)
// =====================================================
// ni, nj, nk, nl: Dimensionen der 4D-Matrix
// bei neutyp=-1 temporaere Matrix erzeugen
// bei neutyp= 1 permanente Matrix erzeugen 
// ----------------------------------------------
MATRIX::MATRIX(int ni,int nj,int nk,int nl,int neutyp)
{
//*** std::cerr<<"Konstruktor: temporaere MATRIX erzeugen, neutyp= "<<neutyp<<'\n'; 
//
this->erzeugen(ni,nj,nk,nl,neutyp);
};//--------------------------------------------------------------------
//
// Kopier-Konstruktor -> MATRIX mat1(&mat2,neutyp)
// ===============================================
// bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
// bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
// bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
// --------------------------------------------------------------------------
MATRIX::MATRIX(MATRIX &mat2,int neutyp)
{
//*** std::cerr<<"MATRIX::Kopier-Konstruktor: Kopie von MATRIX erzeugen, neutyp= "<<neutyp<<'\n';
// 
// permanente Matrix
// =================
this->kopieren(mat2,neutyp);
};//---------------------------------------------------------------------------
//
// Referenztyp wandeln -> MATRIX &mat=referenz(MATRIZEN &mat_)
// ===========================================================
MATRIX &MATRIX::referenz(MATRIZEN &mat)
{
//*** std::cerr<<"MATRIX::referenz &mat= "<<&mat<<" return= "<<&((MATRIX&) mat)<<'\n';
return (MATRIX&) mat;
};//---------------------------------------------------------------------------
//
// Index ijkl von Matrixelement -> mat.index(i,j,k,l)
// ==================================================
int MATRIX::index(int i, int j, int k, int l)
{
char c;
int ijkl;
// Fehler abfangen
// ---------------
ijkl=i+dimi*(j+dimj*(k+dimk*l));
if(ijkl >= dimmax)
{
std::cout<<"*** Fehler in Funktion MATRIX::index -> Falsche Indizes:"<<'\n';
std::cout<<"dimmax="<<dimmax<<'\n';
std::cout<<"dimi="<<dimi<<" dimj="<<dimj<<" dimk="<<dimk<<" diml="<<diml<<'\n';
std::cout<<"   i="<<   i<<"    j="<<   j<<"    k="<<   k<<"    l="<<   l<<'\n';
std::cout<<"  ijkl="<<  ijkl<<'\n';
std::cin>>c;
return 0;
}
//----------------
else
return ijkl;
};//---------------------------------------------------------------------------
//
// 2D-Indizes von Matrizenelement -> mat.index_2d(ij, dimi,dimj, i,j)
// ==================================================================
void MATRIX::index_2d(int ij,int dimi,int dimj,int &i,int &j)
{
int k;
index_3d(ij,dimi,dimj,1, i,j,k);
};//---------------------------------------------------------------------------
//
// 3D-Indizes von Matrizenelement -> mat.index_3d(ijk, dimi,dimj,dimk, i,j,k)
// ==========================================================================
void MATRIX::index_3d(int ijk,int dimi,int dimj,int dimk,int &i,int &j,int &k)
{
char c;
int n2=dimi*dimj;
// ---------------
// k=ijk/(ni*nj)
k=ijk/n2;
// j=ijk/ni-k*nj
n2=k*dimj;
j=ijk/dimi-n2;
// i=ijk-(k*nj+j)*ni
i=ijk-(n2+j)*dimi;
//
// Fehler abfangen
if( (i >= dimi) || (j >= dimj) || (k >= dimk) )
{
std::cout<<" *** Fehler in Funktion MATRIX::index_3d -> Falsche Indizes:"<<'\n';
std::cout<<"ijk= "<<ijk;
std::cout<<"dimi= "<<dimi<<" dimj= "<<dimj<<" dimk= "<<dimk<<'\n';
std::cout<<"   i= "<<   i<<"    j= "<<   j<<"    k= "<<   k<<'\n';
i=j=k=0;
std::cin>>c;
};
};//---------------------------------------------------------------------------
//
// 4D-Indizes von Matrizenelement -> mat.index_4d(ijkl, i,j,k,l)
// =============================================================
void MATRIX::index_4d(int ijkl,int &i,int &j,int &k,int &l)
//
{
char c;
int n2=dimi*dimj;
int n3=n2*dimk;
// ---------------
// l=ijkl/(ni*nj*nk)
l=ijkl/n3;
// k=ijkl/(ni*nj)-l*nk
n3=l*dimk;
k=ijkl/n2-n3;
// j=ijkl/ni-(l*nk+k)*nj
n2=(n3+k)*dimj;
j=ijkl/dimi-n2;
// i=ijkl-((l*nk+k)*nj+j)*ni
i=ijkl-(n2+j)*dimi;
//
// Fehler abfangen
if( (i >= dimi) || (j >= dimj) || (k >= dimk) || (l >= diml) )
{
std::cout<<" *** Fehler in Funktion MATRIX::index_4d -> Falsche Indizes:"<<'\n';
std::cout<<"ijkl= "<<ijkl<<'\n';
std::cout<<"dimi= "<<dimi<<" dimj= "<<dimj<<" dimk= "<<dimk<<" diml= "<<diml<<'\n';
std::cout<<"   i= "<<   i<<"    j= "<<   j<<"    k= "<<   k<<"    l= "<<   l<<'\n';
i=j=k=l=0;
std::cin>>c;
};
};
// ======================================================================================
