// =============== Klassen-Definition: qMATRIX  ========================================
// qMATRIX-Klasse fuer Lineare Gleichungsysteme mit quadratischer Koeffizienten-Matrix
// Quadratmatrix MAT(mk,mk) vom Typ double wird spaltenweise gespeichert !!!
// Indizierung von mk beginnt bei 0 !!!
// vgl. U. Breymann, C++, Carl Hanser Verlag 1993, S.186 ff
// ======================================================================================
// Urheberrechte: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Erkrath/Darmstadt
// experimentelles Lehr- und Studienprogramm - kommerzielle Nutzung nicht gestattet
// Der Autor bernimmt keine Haftung fr Unvollstndigkeiten und Maengel der Algorithmen!
// Version:  2023
// Compiler: C++
// ================ Include-Files =======================================================
#include "pch.h"
#include "framework.h"
#include "MATRIXclass.h"
// ---------------------
// #include "Matrizen.h"
// #include "Matrix.h"
// #include "qMatrix.h"
// ============================================================================
// public -> Oeffentliche Mitgliedsfunktionen
// ============================================================================
//
// Konstruktor -> qMATRIX MAT(mk,symdef)
// =====================================
// mk, mk: Dimensionen der 2D-Matrix
// symdef =  1   symmetrische positiv-definite Koeffizientenmatrix(fr Cholesky-Elimination) 
//        =  0   symmetrische                  Koeffizientenmatrix(fr Gauss-Elimination) 
//        = -1 unsymmetrische Koeffizientenmatrix                 (fr Gauss-Elimination)
// --------------------------------------------------------------------------------------
qMATRIX::qMATRIX(int mk, int symdef) : MATRIX(mk, mk)
{
//*** std::cerr << "qMATRIX::Konstruktor: neue Matrix erzeugen" << '\n';
//
// Matrizentyp
defsym=symdef;
// 
// Speicher fr die gestaffelte Koeffizientenmatrix
O_matrix_staff = NULL;
U_matrix_staff = NULL;
//
// Determinante der gestaffelten Koeffizienten-Matrix
det = 0.0;
//
//*** std::cerr << "qMATRIX: Konstruktor: symdef= " << symdef << " det= " << det << '\n';
};//---------------------------------------------------------------------------
//
// Destruktor -> delete &MAT
// =========================
qMATRIX::~qMATRIX(void)
{
//*** std::cerr << "qMATRIX::Destruktor: symdef= "<< symdef <<'\n';
//
// gestaffelte qMatrix loeschen
if (O_matrix_staff != NULL) { delete[] O_matrix_staff; }
if (U_matrix_staff != NULL) { delete[] U_matrix_staff; }
}; 
//-----------------------------------------------------------------------------
// 
// Matrix zuweisen -> mat1 = mat2 (nur Matrizeninhalt kopieren)
// ==============================
// Hier nochmal, weil der Zuweisungsoperator  = nicht vererbt werden kann !!!
// --------------------------------------------------------------------------
qMATRIX& qMATRIX::operator  = (MATRIX& mat2)
{
    this->zuweisen(mat2, 0);
    //
    // ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
    this->temporaere_matrix_loeschen();
    //
    return *this;
};//---------------------------------------------------------------------------
//
// Matrizendivision    matx(s,n) = matr(i,n) / MATk(i,s)
//         MATk(i,s) * matx(s,n) = matr(i,n)
// =========================================================================================
MATRIZEN& qMATRIX::operator / (qMATRIX& MATk)
{
    // Matrixdimensionen ermitteln
    //*** int ni, nn, nii, ns, dummy;
    //*** this->dimension_ermitteln(ni, nn, dummy, dummy);
    //***  MATk.dimension_ermitteln(nii, ns, dummy, dummy);
    //
    //*** std::cout << "qMATRIX: operator /" << endl;
    //*** std::cout << "qMATRIX::(*this) matr(i,n) / MATk(ii,s) ni, nn, nii, ns = " << ni << nn << nii << ns << '\n';
    //
    // Ergebnisspeicher matx
    MATRIZEN *temp = temporaere_matrix_erzeugen(*this, 1);
    // ------------------------------------------------------
    // 
    // Gleichnugssystem lsen
    *temp = MATk.dividieren(*this);
    // ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
    MATk.temporaere_matrix_loeschen();
    this->temporaere_matrix_loeschen();
    //
    return *temp;
};//========================================================================================
//
// ermittelte Determinante ausgeben -> symdef = MATk.gib_determinante (det)
// symdef = 1 / 0 / -1 fr Cholesky / Gauss (symmetrisch) / Gauss (unsymmetrisch)
// ==============================================================================
    int &qMATRIX::gib_determinante(double &det)
    {
    //*** std::cerr << "qMATRIX: gib_determinante= "<< det << endl;
    //
    det = this->det;
    return this->defsym;
}//==============================================================================
//
//              Private virtuelle Mitgliedsfunktionen
// ==============================================================================
//
// Matrizendivision -> matx(s,n) = matr(i,n) / MATk(i,s)
//         MATk(i,s) * matx(s,n) = matr(i,n)
// =========================================================================================
    MATRIZEN& qMATRIX::dividieren(MATRIZEN& matr_)
    {
        // Referenztyp wandeln
        MATRIX& matr = this->referenz(matr_);
        //
        // Matrixdimensionen ermittel
        int ni, ns, nii, nn, dummy;
        this->dimension_ermitteln(ni,  ns, dummy, dummy);
         matr.dimension_ermitteln(nii, nn, dummy, dummy);
        // 
        //*** std::cout << "qMATRIX::dividieren: matr(i,n) / MATk(i,s) nii, nn, ni, ns = " << nii << nn << ni << ns << '\n';
        //*** std::cout << "qMATRIX::dividieren: defsym= " << this->defsym << endl;
        // ===============================================================================================
        // Koeffizientenmatrix staffeln
        // ============================
        if (O_matrix_staff == NULL)
        {
            // Speicher der gestaffelten Form
            O_matrix_staff = new MATRIX(ni, ns);
            // --------------------------------------------
            // Gestaffelte Form von MATk erzeugen
            if (defsym == 1) det = this->nach_cholesky_staffeln(*O_matrix_staff);
            else
                if (defsym == 0) det = this->nach_gauss_staffeln(*O_matrix_staff);
                else
                {
                    U_matrix_staff = new MATRIX(ni, ns);
                    det = this->nach_gauss_staffeln(*O_matrix_staff, *U_matrix_staff);
                    //
                    //*** std::cout << "qMATRIX: gestaffelte Form U= " << *U_matrix_staff << endl;
                };
            //*** std::cout << "qMATRIX: gestaffelte Form O= " << *O_matrix_staff << endl;
        };
        // ==================================================================================================
        // Rechte Seiten staffeln
        // ======================
        //*** std::cout << "qMATRIX::dividieren: Rechte Seiten= " << matr << endl;
        //
        MATRIZEN* r_staff = temporaere_matrix_erzeugen(matr, 1);
        // 
        if (defsym > -1) matr.staffeln(*O_matrix_staff, *r_staff, defsym);
        else             matr.staffeln(*U_matrix_staff, *r_staff, -1);
        //
        //*** std::cout << "qMATRIX: gestaffelte Rechte Seiten= " << *r_staff << endl;
        // ==================================================================================================
        // Ergebnisspeicher matx
        MATRIZEN* loes = temporaere_matrix_erzeugen(matr, 1);
        // --------------------------------------------------
        // Loesungen ermitteln
        // ===================
        *loes = matr.loesungen(*O_matrix_staff, *r_staff, defsym);
        //*** std::cout << "qMATRIX::dividieren Loesungen= " << *loes << endl;
        // ---------------------------------------------------------------------
        // ggfls. temporaere Matrix loeschen, Original-Deskriptor wiederherstellen
        r_staff->temporaere_matrix_loeschen();
        matr_.temporaere_matrix_loeschen();
        this->temporaere_matrix_loeschen();
        //
        return *loes;
    }//============================================================================